{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "puzzle_input_str = open(\"puzzle_input/day16.txt\").read()\n",
    "\n",
    "def parse_sample(sample_str):\n",
    "    before_str, instr_str, after_str = sample_str.split(\"\\n\")\n",
    "\n",
    "    before = [int(s) for s in before_str[9:][:-1].split(\", \")]\n",
    "    instr = [int(s) for s in instr_str.split(\" \")]\n",
    "    after = [int(s) for s in after_str[9:][:-1].split(\", \")]\n",
    "    \n",
    "    return (before, instr, after)\n",
    "\n",
    "def parse_samples(samples_str):\n",
    "    return [ parse_sample(sample_str) for sample_str in samples_str.split(\"\\n\\n\")]\n",
    "\n",
    "def parse_program(program_str):\n",
    "    return program_str.split(\"\\n\")\n",
    "\n",
    "def parse_input(input_str):\n",
    "    samples_str, puzzle_program_str = input_str.split(\"\\n\\n\\n\\n\")\n",
    "    \n",
    "    return parse_samples(samples_str), parse_program(puzzle_program_str)\n",
    "    \n",
    "samples, program = parse_input(puzzle_input_str)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "class VM(object):\n",
    "    def __init__(self, register_count):\n",
    "        self.register = [0,0,0,0]\n",
    "        self.opcodes = {\n",
    "            \"addr\": self.addr,\n",
    "            \"addi\": self.addi,\n",
    "            \"mulr\": self.mulr,\n",
    "            \"muli\": self.muli,\n",
    "            \"banr\": self.banr,\n",
    "            \"bani\": self.bani,\n",
    "            \"borr\": self.borr,\n",
    "            \"bori\": self.bori,\n",
    "            \"setr\": self.setr,\n",
    "            \"seti\": self.seti,\n",
    "            \"gtir\": self.gtir,\n",
    "            \"gtri\": self.gtri,\n",
    "            \"gtrr\": self.gtrr,\n",
    "            \"eqir\": self.eqir,\n",
    "            \"eqri\": self.eqri,\n",
    "            \"eqrr\": self.eqrr\n",
    "        }\n",
    "\n",
    "        \n",
    "    def execute(self, opcode, a, b, c):\n",
    "        decode = {\n",
    "            0 : self.eqir,\n",
    "            1 : self.seti,\n",
    "            2 : self.eqri,\n",
    "            3 : self.eqrr,\n",
    "            4 : self.addi,\n",
    "            5 : self.setr,\n",
    "            6 : self.gtrr,\n",
    "            7 : self.gtri,\n",
    "            8 : self.muli,\n",
    "            9 : self.bori,\n",
    "            10: self.bani,\n",
    "            11: self.borr,\n",
    "            12: self.gtir,\n",
    "            13: self.banr,\n",
    "            14: self.addr,\n",
    "            15: self.mulr\n",
    "        }\n",
    "        \n",
    "        assert(decode[opcode](a,b,c))        \n",
    "        \n",
    "    def any_invalid_registers(self, regs):\n",
    "        result = True\n",
    "\n",
    "        for reg in regs:\n",
    "            result = result & (reg < len(self.register))\n",
    "\n",
    "        return not result\n",
    "\n",
    "    def test_opcode(self, opcode_name, initial_register_state, a, b, c, final_register_state):\n",
    "        self.register = [ reg for reg in initial_register_state ]\n",
    "        result = self.opcodes[opcode_name](a, b, c)\n",
    "        return result and (self.register == final_register_state)\n",
    "\n",
    "    # addr (add register) stores into register C the result of adding register A and register B.\n",
    "    def addr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a,b,c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] + self.register[b]\n",
    "        return True\n",
    "\n",
    "    # addi (add immediate) stores into register C the result of adding register A and value B.\n",
    "    def addi(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False    \n",
    "\n",
    "        self.register[c] = self.register[a] + b\n",
    "        return True\n",
    "\n",
    "    # mulr (multiply register) stores into register C the result of multiplying register A and register B.\n",
    "    def mulr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] * self.register[b]\n",
    "        return True\n",
    "\n",
    "    # muli (multiply immediate) stores into register C the result of multiplying register A and value B.\n",
    "    def muli(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] * b\n",
    "        return True\n",
    "\n",
    "    # banr (bitwise AND register) stores into register C the result of the bitwise AND of register A and register B.\n",
    "    def banr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] & self.register[b]\n",
    "        return True\n",
    "\n",
    "    # bani (bitwise AND immediate) stores into register C the result of the bitwise AND of register A and value B.\n",
    "    def bani(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] & b\n",
    "        return True\n",
    "\n",
    "    # borr (bitwise OR register) stores into register C the result of the bitwise OR of register A and register B.\n",
    "    def borr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] | self.register[b]\n",
    "        return True\n",
    "\n",
    "\n",
    "    # bori (bitwise OR immediate) stores into register C the result of the bitwise OR of register A and value B.\n",
    "    def bori(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a] | b\n",
    "        return True\n",
    "\n",
    "    # setr (set register) copies the contents of register A into register C. (Input B is ignored.)\n",
    "    def setr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = self.register[a]\n",
    "        return True\n",
    "\n",
    "    # seti (set immediate) stores value A into register C. (Input B is ignored.)\n",
    "    def seti(self, a, b, c):\n",
    "        if self.any_invalid_registers([c]):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = a\n",
    "        return True\n",
    "\n",
    "    # gtir (greater-than immediate/register) sets register C to 1 if value A is greater than register B. Otherwise, register C is set to 0.\n",
    "    def gtir(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, b)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(a > self.register[b])\n",
    "        return True\n",
    "\n",
    "    # gtri (greater-than register/immediate) sets register C to 1 if register A is greater than value B. Otherwise, register C is set to 0.\n",
    "    def gtri(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, a)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] > b)\n",
    "        return True\n",
    "\n",
    "    # gtrr (greater-than register/register) sets register C to 1 if register A is greater than register B. Otherwise, register C is set to 0.\n",
    "    def gtrr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a, b)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] > self.register[b])\n",
    "        return True\n",
    "\n",
    "    # eqir (equal immediate/register) sets register C to 1 if value A is equal to register B. Otherwise, register C is set to 0.\n",
    "    def eqir(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, b)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(a == self.register[b])\n",
    "        return True\n",
    "\n",
    "    # eqri (equal register/immediate) sets register C to 1 if register A is equal to value B. Otherwise, register C is set to 0.\n",
    "    def eqri(self, a, b, c):\n",
    "        if self.any_invalid_registers((c, a)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] == b)\n",
    "        return True\n",
    "\n",
    "    # eqrr (equal register/register) sets register C to 1 if register A is equal to register B. Otherwise, register C is set to 0.\n",
    "    def eqrr(self, a, b, c):\n",
    "        if self.any_invalid_registers((a,b,c)):\n",
    "            return False\n",
    "\n",
    "        self.register[c] = int(self.register[a] == self.register[b])\n",
    "        return True"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "521\n"
     ]
    }
   ],
   "source": [
    "# Part 1\n",
    "\n",
    "from pprint import pprint \n",
    "\n",
    "vm = VM(4)\n",
    "\n",
    "matching_opcode_by_name = { opcode_name:set() for opcode_name in vm.opcodes }\n",
    "matching_opcode_by_code = [ set() for opcode_name in vm.opcodes ]\n",
    "little_test_data =[\n",
    "    [ [0, 2, 2, 2], 4, 2, 3, 2, [0, 2, 5, 2] ],\n",
    "#     [ [3, 2, 1, 1], 9, 2, 1, 2, [3, 2, 2, 1] ]\n",
    "]\n",
    "\n",
    "three_or_more_matching = 0\n",
    "\n",
    "for test_case in samples:\n",
    "    before, (op, a, b, c), after = test_case\n",
    "    matching_opcodes = 0\n",
    "    for opcode_name in vm.opcodes:\n",
    "        if vm.test_opcode(opcode_name, before, a, b, c, after):\n",
    "            matching_opcode_by_name[opcode_name].add(op)\n",
    "            matching_opcode_by_code[op].add(opcode_name)\n",
    "            matching_opcodes += 1\n",
    "    if (matching_opcodes >= 3):\n",
    "        three_or_more_matching += 1\n",
    "    \n",
    "print(three_or_more_matching)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "{'addi': {1, 4, 5, 8, 9, 10, 13, 14, 15},\n",
      " 'addr': {0, 1, 5, 9, 10, 11, 14},\n",
      " 'bani': {10},\n",
      " 'banr': {10, 13},\n",
      " 'bori': {1, 5, 9, 10, 13},\n",
      " 'borr': {0, 1, 5, 9, 10, 11, 13},\n",
      " 'eqir': {0, 12, 13, 6},\n",
      " 'eqri': {0, 2, 12, 6},\n",
      " 'eqrr': {2, 3, 12, 13},\n",
      " 'gtir': {10, 12, 5},\n",
      " 'gtri': {3, 5, 6, 7, 10, 12, 13},\n",
      " 'gtrr': {10, 12, 5, 6},\n",
      " 'muli': {8, 9, 10, 14},\n",
      " 'mulr': {1, 8, 9, 10, 11, 13, 14, 15},\n",
      " 'seti': {0, 1, 2, 3, 5, 6, 7, 10, 12, 13},\n",
      " 'setr': {10, 13, 5}}\n",
      "0 borr, eqri, eqir, seti, addr\n",
      "1 mulr, borr, bori, seti, addi, addr\n",
      "2 eqri, eqrr, seti\n",
      "3 eqrr, seti, gtri\n",
      "4 addi\n",
      "5 gtrr, borr, bori, setr, gtir, seti, gtri, addi, addr\n",
      "6 gtrr, eqri, eqir, seti, gtri\n",
      "7 seti, gtri\n",
      "8 mulr, addi, muli\n",
      "9 mulr, borr, bori, addi, muli, addr\n",
      "10 mulr, gtrr, borr, bori, setr, bani, banr, gtir, seti, gtri, addi, muli, addr\n",
      "11 mulr, borr, addr\n",
      "12 gtrr, eqri, gtir, eqir, seti, eqrr, gtri\n",
      "13 mulr, borr, bori, setr, banr, eqir, seti, gtri, eqrr, addi\n",
      "14 mulr, addi, muli, addr\n",
      "15 mulr, addi\n"
     ]
    }
   ],
   "source": [
    "pprint(matching_opcode_by_name)\n",
    "\n",
    "for i, opcodes in enumerate(matching_opcode_by_code):\n",
    "    print(i, \", \".join(opcodes))\n",
    "\n",
    "# mapping is:\n",
    "#         0  eqir\n",
    "#         1  seti\n",
    "#         2  eqri\n",
    "#         3  eqrr\n",
    "#         4  addi\n",
    "#         5  setr\n",
    "#         6  gtrr\n",
    "#         7  gtri\n",
    "#         8  muli\n",
    "#         9  bori\n",
    "#         10 bani\n",
    "#         11 borr\n",
    "#         12 gtir\n",
    "#         13 banr\n",
    "#         14 addr\n",
    "#         15 mulr"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "594\n"
     ]
    }
   ],
   "source": [
    "vm = VM(4)\n",
    "vm.register = [0,0,0,0]\n",
    "\n",
    "for raw_instruction in program:\n",
    "    op, a, b, c = [int(s) for s in raw_instruction.split(\" \")]\n",
    "    vm.execute(op, a, b, c)\n",
    "    \n",
    "print(vm.register[0])"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
